
/*

	Animated Limited Region
	v1.0
	Matt Gorner
	www.pixsim.co.uk

	History:
	
	1.0 (22 Dec 2010) Initial release

	----------------------------------------------------------------------------------------------------------

	Description:
	
	Allows animation of the Limited Region tool in Layout.
	
	Usage:
	
	1) Add as Master Script (CTRL q or go to 'Ultilities' tab > Plugins > Master Plugins)
	
	2) Grab current Limited Region (turns it on if off)
	
	3) Set whether you want with or without Borders

	4) Use handles to adjust (grab again to update the list)

	5) Turn on 'Interpolate' mode if you want to tween (linearly) between stored frames, or set to off to just render the stored frames as is

	Notes:

	Because this script renders using an F9 sequence which does NOT invoke saving via the 'Render Options' > 'Output' tab,
	you need to set image saving via the 'Effects' panel > 'Processing' tab (CTRL F8) under 'Add Image Filter'.
	
	If you don't want the 'Image Viewer' to show up with each render, I recommened you download Mike Wolf's excellent 'db&w Tools' plugin (http://www.db-w.com/content/view/137/175/)
	This contains an Image Viewer called 'Revenge of the Lenscap' which basically does nothing!  As setting the 'Render Display' to 'None' will still invoke the 'Turn On Image Viewer?' requester,
	which will stop the render from progressing.
	
	Acknowledgements:
	
	Thanks to Jon Tindall for helping me with figuring out the linear interpolation math! :)

*/


@version 2.2
@warnings
@script master
@name "Animated Limited Region"

@define SCRIPT_NAME    "Animated Limited Region"
@define SCRIPT_VERSION "1.0"
@define DEBUG 0

// Global Variables

rendering;
alertLevel;
first_frame, last_frame, rendering_frame, currentframe;
//rX, rY, rW, rH, border;
framelistdisplay, selected_entry, just_selecting;
host;
interp;
list_position;

//Gadgets
gad_first, gad_last, gad_step, gad_grab, gad_use_borders, gad_interpolate;

// One-time initialization takes place here

create
{
	// Init variables
	just_selecting	= FALSE;
    rendering		= FALSE;
    interp			= TRUE;
    host			= runningUnder();

	first_frame		= get_first_frame();
	last_frame		= get_last_frame();
	currentframe	= round(Scene().currenttime * Scene().fps,1);
	rendering_frame	= first_frame;
	list_position	= 1;

	setDescription();
}


// Procedure to update script name in Master Plugins to show frame sequence
setDescription
{
	setdesc( SCRIPT_NAME + " " + SCRIPT_VERSION );
}

flags
{
    // Tell LightWave what type of master script level we are:
    // LWMASTF_LAYOUT	= Layout level script, these aren't removed if the scene is cleared (we also NEED to use this one or the script doesn't work)
	// LWMAST_SCENE		= Scene level script, these will be removed if the scene is cleared

    // Also tell it we want to recieve notifications:
    // LWMASTF_RECEIVE_NOTIFICATIONS

    // These flags definitions can be found in the LW SDK Include file: 'lwmaster.h'

    // return(LWMASTF_LAYOUT,LWMASTF_RECEIVE_NOTIFICATIONS);
    return(SCENE,LWMASTF_RECEIVE_NOTIFICATIONS);
}

options
{
	// Is the requester open already?  Then close it.
    if(reqisopen())
        reqend();
    else
    {
    	// Otherwise open it up ...

	    // User Interface Layout Variables (this is me being anal!) :D

		gad_x				= 10;										// Gadget X coord
		gad_y				= 10;										// Gadget Y coord
		gad_w				= 260;										// Gadget width
		gad_h				= 19;										// Gadget height
		gad_text_offset		= 60;										// Gadget text offset
		num_gads			= 4;										// Total number of gadgets vertically (for calculating the max window height)
		ui_spacing			= 4;										// Spacing gap size
		ui_offset_x 		= 0;										// Main X offset from 0
		ui_offset_y 		= ui_spacing*2;								// Main Y offset from 0
		ui_row				= 0;										// Row number
		ui_row_offset		= gad_h + ui_spacing;						// Row offset
		ui_window_w			= 400;										// Window width
		ui_window_h			= 224;										// Window height

		reqbegin(SCRIPT_NAME + " " + SCRIPT_VERSION);
		reqsize(ui_window_w,ui_window_h);

		// Define and layout gadgets ...

		gad_frame_list = ctllistbox("Frame, X, Y, X2, Y2, Border",300,300,"lb_count","lb_name","lb_event");
		ctlposition(gad_frame_list, gad_x, gad_y, ui_window_w - (2 * gad_x + 22), ui_window_h - gad_h - 74);
		ui_offset_y += ui_row_offset + 120;

    	gad_grab = ctlbutton("Grab Current Region",0,"grab_button");
    	ctlposition(gad_grab, gad_x - 60, ui_offset_y, gad_w - 30, gad_h, gad_text_offset);

    	ui_offset_y += ui_row_offset;

		gad_use_borders = ctlcheckbox("With Borders?", 0);
		ctlposition(gad_use_borders, gad_x - 60, ui_offset_y, gad_w - 30, gad_h, gad_text_offset);
		ctlrefresh(gad_use_borders, "update_use_borders");

   		ui_offset_y += ui_row_offset;

    	gad_remove = ctlbutton("Remove",0,"remove_button");
    	ctlposition(gad_remove, gad_x - 60, ui_offset_y, gad_w - 117, gad_h, gad_text_offset);

    	gad_clear = ctlbutton("Clear",0,"clear_button");
    	ctlposition(gad_clear, gad_x + 27, ui_offset_y, gad_w - 117, gad_h, gad_text_offset);

   		ui_offset_y -= ui_row_offset;
   		ui_offset_y -= ui_row_offset;

		gad_x = 190;

    	gad_interpolate = ctlcheckbox("Interpolate", interp);
    	ctlposition(gad_interpolate, gad_x - 30, ui_offset_y, gad_w - 30, gad_h, gad_text_offset);
    	ctlrefresh(gad_interpolate, "update_interpolate");
    	ui_offset_y += ui_row_offset;
    	ui_offset_y += ui_row_offset;

    	gad_button = ctlbutton("Render F9 Sequence",0,"render_button");
    	ctlposition(gad_button, gad_x - 30, ui_offset_y, gad_w - 30, gad_h, gad_text_offset);

    	reqopen();
    }
}

// Save frames to Scene file ...
save: mode, io
{
    return if mode == OBJECTMODE;

    io.writeln(framelistdisplay.size().asStr()); // Number of entries

    for(x = 1;x <= framelistdisplay.size();x++)
        io.writeln(framelistdisplay[x]);
}

// Load frames from Scene file ...
load: mode,io
{
    return if mode == OBJECTMODE;

    index = 1;

    count = io.read().asInt();
    while(count)
    {
		framelistdisplay[index] = io.read();
		++index;
		--count;
	}
	
	// Set variables (just in case)

	first_frame		= get_first_frame();
	last_frame		= get_last_frame();
	currentframe	= first_frame;
	rendering_frame	= first_frame;
	list_position	= 1;

	// Open up UI on load (just to remind the user that it's saved within a scene)
	// Comment out the next line if you don't want this behaviour
	options();
}


// LightWave event processing ...
process: event, command
{
	// Won't work with Screamernet as it needs F9 rendering
    return if host == SCREAMERNET;

	// We got an event passed, has the render been aborted?
    if(event == LWEVNT_NOTIFY_RENDER_ABORTED)
    {
    	// Apparently so, so update our script variable to reflect we are no longer rendering
        rendering		= FALSE;
        frame_to_render = get_first_frame();
        list_postiion	= 1;

        // Re-enable dialog boxes and user alert level
        AutoConfirm(0);
        AlertLevel(alertLevel);
    }

	// The timeline slider was moved ...
    if( event == LWEVNT_NOTIFY_MOTION_UPDATE_COMPLETE && rendering == FALSE )
    {
    	// Update the Limited Region
		set_limited_region();
    }

	// We got a render complete event (Note: LWEVNT_RENDER_DONE does not work with this script, we need to use this event)
    if(event == LWEVNT_NOTIFY_RENDER_COMPLETE )
    {
	    if (framelistdisplay != NULL)
	    {
	    	// Are we still rendering?  And is the frame to render less or equal to the last frame we want to render?
	    	if (rendering == true && ( rendering_frame < last_frame ))
	    	{
	    		// Interpolate mode active?
				if(interp)
				{
					// Yes, just increment the frame then
					rendering_frame ++;
				}else
				{
					// Nope, increment the list position so we can extract the correct frame number stored
					list_position	++;
					s = parse(",", framelistdisplay[ list_position ]);
					rendering_frame = s[1].asInt();
				}
				
				// Set rendering going again ...
				render();

			}else
			{
				// Nope, we're finished, reset the variables
				rendering		= FALSE;
				rendering_frame = get_first_frame();
				list_position	= 1;

	        	// Re-enable dialog boxes and user alert level
	            AutoConfirm(0);
	            AlertLevel(alertLevel);
			}

		} // framelistdisplay

    } // event

}

// Gadget refresh & callback procedures ...

// Keep track of how many items are in the ctllistbox gadget
lb_count
{
	return(framelistdisplay? framelistdisplay.count() : 0);
}

// The list of names that show up in the ctllistbox gadget
lb_name: index
{
	return( framelistdisplay[ index ] );
}

// Which one was selected?
lb_event: index
{
	if(index)
	{
		selected_entry = index;
		just_selecting = TRUE;

		s = parse(",", framelistdisplay[ selected_entry[1] ]);
	
	 	if(s[s.count()] == " No Border")
		{
			setvalue(gad_use_borders, 0);
		}else
		{
			setvalue(gad_use_borders, 1);
		}

		currentframe = s[1].asNum();
	
		requpdate();
		update_layout();

		CommandInput(s[6] == " With Border"? "LimitedRegion 1" : "LimitedRegion 2");
		RegionPosition(s[2].asNum() , s[3].asNum() , s[4].asNum() - s[2].asNum() , s[5].asNum() - s[3].asNum());
	}
}

// Borders checkbox button callback
update_use_borders: value
{
	if( framelistdisplay == NULL) return;

	str = NULL;

    if(selected_entry == nil)
        return;

    if(just_selecting == TRUE)
    {
		CommandInput(value == 1? "LimitedRegion 1" : "LimitedRegion 2");
    	just_selecting = FALSE;
        return;
    }

	s = parse(",", framelistdisplay[ selected_entry[1] ]);

	str = string(s[1], "," , s[2] , "," , s[3] ,  "," , s[4] , "," , s[5] , "," , value == 0? " No Border" : " With Border");

	framelistdisplay[ selected_entry[1] ] = str;

	CommandInput(value == 1? "LimitedRegion 1" : "LimitedRegion 2");
	requpdate();
}

// Interpolate checkbox button callback
update_interpolate: value
{
	interp = value;
}

// Render button callback
render_button
{
    // Suppress dialog boxes
    AutoConfirm(1);

	// Store the current user alert level
    alertLevel = Scene().alertlevel;

    // Set the Alert Level to expert (silence them basically)
    AlertLevel(ALERT_EXPERT);

	// Yes we will be rendering ...
    rendering = true;
    rendering_frame = get_first_frame();
    //update_layout();
    //set_limited_region();

    // Got to render procedure
    render();
}

// Grab button callback
grab_button
{
    scene = Scene();

	// Are Fractional Frames active?  Tell the user to turn them off, they're not supported yet, unfortunately.
	if(scene.generalopts[4])
	{
		warn("Sorry, Fractional Frames are not currently supported, please turn them off.");
		return;
	}

    // If Limited Region is not switched on, turn it on
    if (!scene.renderopts[8])
	{
		LimitedRegion();
	}

	frame	= round(scene.currenttime * scene.fps,1);

    lrx			= scene.limitedregion[1];
    lry			= scene.limitedregion[2];
    lrx2		= scene.limitedregion[3] + 1; // Need to add 1 to stop ever shrinking box when clicked multiple times!
    lry2		= scene.limitedregion[4] + 1; // Need to add 1 to stop ever shrinking box when clicked multiple times!

	// If Camera has Limited Region on
    if(Camera().flags[3])
	{
 		if (!Camera().flags[4]) // LR with borders
		{
			str = string(frame , "," , lrx , "," , lry , "," , lrx2 , "," , lry2 , ", With Border");
		}else // LR without borders
		{
			str = string(frame , "," , lrx , "," , lry , "," , lrx2 , "," , lry2 , ", No Border");
		}
	}

	// Does the frame we're grabbing already exist in the list?
	addto = frame_exists(frame);

	// Yes it does, so replace it
	if( addto != -1 )
	{
		// Add to list
		framelistdisplay[ addto ] = str;
	}else // No, so add new frame to list
	{
		// Add to list
		framelistdisplay += str;
	}

	sort_frames();

	// Update globals
	first_frame		= get_first_frame();
	last_frame		= get_last_frame();
	rendering_frame = first_frame;

	// Update requester
	requpdate();
}

remove_button
{
    if(selected_entry == nil)
        return;

	// Build temp array
    t = nil;
    for(x = 1;x <= framelistdisplay.count();x++)
    {
        if(!selected_entry.contains(x))
            t += framelistdisplay[x];
    }

	// Assign it to global
    framelistdisplay = t;

	// Update globals
	first_frame		= get_first_frame();
	last_frame		= get_last_frame();
	rendering_frame = first_frame;
	
	// Update requester
    requpdate();
}

clear_button
{
	// Update globals
	framelistdisplay	= nil;
	first_frame			= nil;
	last_frame			= nil;
	rendering_frame 	= nil;

	// Update requester
    requpdate();
}



// Utility procedures ...

// Simple bubble sort routine
sort_frames
{
	for(x = 1; x < framelistdisplay.size(); x++)
	{
		for(y = 1; y < framelistdisplay.size(); y++)
		{
			s = parse(",", framelistdisplay[ y ]);
			t = parse(",", framelistdisplay[ y + 1 ]);
			if(s[1].asNum() > t[1].asNum())
			{
	        	temp					= framelistdisplay[y + 1];
	        	framelistdisplay[y + 1]	= framelistdisplay[y];
	        	framelistdisplay[y]		= temp;
			}
			s = parse(",", framelistdisplay[ y ]);
			t = parse(",", framelistdisplay[ y + 1 ]);
		}
	}
}

update_layout
{
	// Have to do it this way as GoToFrame( currentframe ); crashes for some reason!
	str = "GoToFrame " + currentframe.asStr();
	CommandInput(str);
}

get_first_frame
{
	if( framelistdisplay == NULL) return(NULL);

	s = parse(",", framelistdisplay[ 1 ]);

	return(s[1].asInt());
}

get_last_frame
{
	if( framelistdisplay == NULL) return(NULL);

	s = parse(",", framelistdisplay[ framelistdisplay.count() ]);

	return(s[1].asInt());
}


// Render procedure
render
{
	if( framelistdisplay == NULL) return;

	// Interpolate mode active?
	if(interp)
	{
		info("Interpolation Mode  |  Rendering Frame: ", rendering_frame, "  |  X ", rX, "   Y ", rY, "   W ", rW, "   H ", rH, "   ", border);
	}else
	{
		s = parse(",", framelistdisplay[ list_position ]);
		info("Rendering Frame: ", s[1] , "  |  X ", rX, "   Y ", rY, "   W ", rW, "   H ", rH, "   ", border);
	}

	GoToFrame( rendering_frame );

	set_limited_region();

	// Render the frame (equivalent of doing an F9)
    CommandInput("RenderFrame");
}

frame_exists: frame
{
	s = NULL;

	// Step through each entry in the list
	for(loop = 1;loop <= framelistdisplay.size(); loop++)
    {
    	// Parse the entry at the loop number
		s = parse(",", framelistdisplay[ loop ]);

		// Does the entry match the frame we're looking for?
		if (s[1].asNum() == frame.asNum() )
		{
			// Yes, then return the position in the array where the match was found
			return( loop );
		}
    }

    // Otherwise return -1
    return( -1 );
}

get_previous_keyframe: current_frame
{
	// Get first entry in the Limited Region Keyframe list
	f = parse(",", framelistdisplay[ 1 ]);

	// Loop backwards from frame passed to frame first in the list (which will always be the lowest)
	for(loop = current_frame - 1; loop >= f[1].asNum(); loop--)
	{
		// Check if a keyframe exist at the frame number?
		w = frame_exists( loop );
	   	if (w != -1)
	   	{
			// Return the frame number
	   		return(w);
	   	}
	}

	// We didn't find one, so return NULL
	return(NULL);
}

get_next_keyframe: current_frame
{
	// Get last entry in the Limited Region Keyframe list
	f = parse(",", framelistdisplay[ framelistdisplay.size() ]);

	// Loop from frame passed to frame last in the list (which will always be the highest)
	for(loop = current_frame + 1; loop <= f[1].asNum(); loop++)
	{
		// Check if a keyframe exist at the frame number?
		w = frame_exists( loop );
	   	if (w != -1)
	   	{
	   		// Return the frame number
	   		return(w);
	   	}
	}

	// We didn't find one, so return NULL
	return(NULL);
}

linear_interpolation: a, b, t
{
	x = a.asNum() + (b.asNum() - a.asNum()) * t.asNum();
	return(x.asNum());
}

set_limited_region
{
	// Don't do anything if there are no keyframes stored by the user
	if( framelistdisplay != NULL)
	{
		scene = Scene();

		// Where are we on the slider?
		currentframe = round(scene.currenttime * scene.fps,1);

		// Does a stored Limited Region key frame exist?
		exists = frame_exists(currentframe);
	
		if(exists != -1)
		{
			// Yes, there is, so just call up the stored values, no need to interpolate
			p = parse(",", framelistdisplay[ exists ]);
	
			currF	= p[1].asNum();
	    	rX		= p[2].asNum();
	    	rY		= p[3].asNum();
	    	rW		= p[4].asNum();
	    	rH		= p[5].asNum();
	    	border	= p[6];
	
			// Set the Border property
			CommandInput(p[6] == " No Border"? "LimitedRegion 2" : "LimitedRegion 1");
	
			// Set the Region position
			RegionPosition(rX, rY, rW - rX, rH - rY);

		}else
		{
			// No, we're on a normal frame, do we need to interpolate?
			if(interp)
			{
				// Apparently so ...
	
				// Get the previous and next Limited Region keyframes in relation to the frame we're currently on (functions return position in framelistdisplay[])
				prv = get_previous_keyframe(currentframe);
				nxt = get_next_keyframe(currentframe);

				if (prv)
				{
	 				p		= parse(",", framelistdisplay[ prv ]);
			    	prev_fr	= p[1].asNum();	// Frame
			    	prev_x1	= p[2].asNum();	// x1
			    	prev_y1	= p[3].asNum();	// y1
			    	prev_x2	= p[4].asNum();	// x2
			    	prev_y2	= p[5].asNum();	// y2
			    	prev_bd = p[6];			// Border string
		    	} // nxt
	
				if (nxt)
				{
					n		= parse(",", framelistdisplay[ nxt ]);
			    	next_fr	= n[1].asNum();	// Frame
			    	next_x1	= n[2].asNum();	// x1
			    	next_y1	= n[3].asNum();	// y1
			    	next_x2	= n[4].asNum();	// x2
			    	next_y2	= n[5].asNum();	// y2
			    	next_bd = n[6];			// Border string
			    } // prv
	
				if(prv)
				{
					// Set the Border property
					CommandInput(prev_bd == " No Border"? "LimitedRegion 2" : "LimitedRegion 1");
					border = prev_bd;
				}else if(nxt)
				{
					// Set the Border property
					CommandInput(nxt_bd == " No Border"? "LimitedRegion 2" : "LimitedRegion 1");
					border = nxt_bd;
				}
	
				// Are we between keyframes?
				if(prv && nxt)
				{
					t	= (currentframe.asNum() - prev_fr.asNum()) / (next_fr.asNum() - prev_fr.asNum());
	
					rX = linear_interpolation( prev_x1, next_x1, t);
					rY = linear_interpolation( prev_y1, next_y1, t);
					rW = linear_interpolation( prev_x2, next_x2, t);
					rH = linear_interpolation( prev_y2, next_y2, t);
	
					// Set the interpolated Region position
					RegionPosition(rX, rY, rW - rX, rH - rY);
				}

			} // interp
	
		} // if exists
	
	} // framelistdisplay

}

DBinfo: str
{
	if(DEBUG == 0)
		return;

	info(str);
}